import logging
import os
import shlex
import telnetlib
import time
from typing import AnyStr, Optional

from pytest_embedded.log import DuplicateStdoutPopen
from pytest_embedded.utils import to_bytes, to_str


class OpenOcd(DuplicateStdoutPopen):
    """
    Class to communicate to OpenOCD
    """

    SOURCE = 'OPENOCD'
    REDIRECT_CLS = None

    OPENOCD_PROG_PATH = 'openocd'
    OPENOCD_DEFAULT_ARGS = '-f board/esp32-wrover-kit-3.3v.cfg'

    TCL_BASE_PORT = 6666
    TELNET_BASE_PORT = 4444
    GDB_BASE_PORT = 3333

    def __init__(
        self,
        openocd_prog_path: Optional[str] = None,
        openocd_cli_args: Optional[str] = None,
        port_offset: int = 0,
        **kwargs,
    ):
        openocd_prog_path = openocd_prog_path or os.getenv('OPENOCD_BIN', self.OPENOCD_PROG_PATH)
        openocd_cli_args = shlex.split(openocd_cli_args or self.OPENOCD_DEFAULT_ARGS)

        openocd_scripts_path = os.getenv('OPENOCD_SCRIPTS')
        if openocd_scripts_path:
            openocd_cli_args.extend(['-s', openocd_scripts_path])

        self.tcl_port = self.TCL_BASE_PORT + port_offset
        self.telnet_port = self.TELNET_BASE_PORT + port_offset
        self.gdb_port = self.GDB_BASE_PORT + port_offset

        openocd_cli_args.extend([
            '-c',
            f'tcl_port {self.tcl_port}',
            '-c',
            f'telnet_port {self.telnet_port}',
            '-c',
            f'gdb_port {self.gdb_port}',
        ])

        super().__init__(cmd=[openocd_prog_path, *openocd_cli_args], **kwargs)

        # open telnet port to interact with openocd
        for i in range(30):
            try:
                self.telnet = telnetlib.Telnet('127.0.0.1', self.telnet_port, 5)
                break
            except ConnectionRefusedError:
                time.sleep(1)
        else:
            raise ConnectionRefusedError

    def write(self, s: AnyStr) -> str:
        # read all output already sent
        resp = self.telnet.read_very_eager()
        if resp:
            logging.debug(f'{self.SOURCE} <-: {to_str(resp)}')

        logging.debug(f'{self.SOURCE} ->: {to_str(s)}')
        self.telnet.write(to_bytes(s, '\n'))

        resp = self.telnet.read_until(b'>')

        logging.debug(f'{self.SOURCE} <-: {to_str(resp)}')
        return to_str(resp)



import os

import pytest

jtag_connection_required = pytest.mark.skipif(
    os.getenv('DONT_SKIP_JTAG_TESTS', False) is False,
    reason='Connect the board to a JTAG adapter then ' 'use "DONT_SKIP_JTAG_TESTS" to run this test.',
)


@jtag_connection_required
def test_pexpect_by_jtag(testdir):
    testdir.makepyfile(r"""
        import os
        import time

        def test_pexpect_by_jtag(dut):
            dut.gdb.write('mon reset halt')
            dut.gdb.write('thb app_main')
            dut.gdb.write('c')
            dut.gdb.write('c', non_blocking=True)
            dut.expect('Hello world')
    """)

    result = testdir.runpytest(
        '-s',
        '--embedded-services', 'jtag,idf',
        '--app-path', os.path.join(testdir.tmpdir, 'hello_world_esp32'),
        '--port', '/dev/ttyUSB1',
        '--part-tool', os.path.join(testdir.tmpdir, 'gen_esp32part.py'),
    )

    result.assert_outcomes(passed=1)



import logging
import re
import shlex
import time
from typing import AnyStr, Optional

from pytest_embedded.log import DuplicateStdoutPopen


class Gdb(DuplicateStdoutPopen):
    SOURCE = 'GDB'
    REDIRECT_CLS = None

    GDB_PROG_PATH = 'xtensa-esp32-elf-gdb'
    GDB_DEFAULT_ARGS = '--quiet'

    _GDB_RESPONSE_FINISHED_RE = re.compile(r'^\(gdb\)\s*$')

    def __init__(self, gdb_prog_path: Optional[str] = None, gdb_cli_args: Optional[str] = None, **kwargs):
        gdb_prog_path = gdb_prog_path or self.GDB_PROG_PATH
        gdb_cli_args = shlex.split(gdb_cli_args or self.GDB_DEFAULT_ARGS)

        self._gdb_first_prompt_matched = False

        super().__init__(cmd=[gdb_prog_path, *gdb_cli_args], **kwargs)

    def write(self, s: AnyStr, non_blocking: bool = False, timeout: float = 30) -> Optional[str]:
        super().write(s)
        _buffer = ''
        _t_start = time.time()

        with open(self._logfile) as fr:
            fr.seek(self._logfile_offset)
            while True:
                line = fr.readline()
                _t_now = time.time()
                if non_blocking:
                    logging.debug('non-blocking write...')
                    return None

                if (_t_now - _t_start) >= timeout:
                    logging.debug(f'current buffer: {_buffer}')
                    raise ValueError(f'gdb no response after {timeout} seconds')

                if line:
                    _buffer += line
                    if self._GDB_RESPONSE_FINISHED_RE.match(line):
                        if not self._gdb_first_prompt_matched:
                            self._gdb_first_prompt_matched = True
                            continue
                        break

            self._logfile_offset = fr.tell()

        logging.debug(f'{self.SOURCE} <-: {_buffer}')
        return _buffer
